__version__ = "0.6.15"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.6.15)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '5339eb78dce37b619d7d1f0e26c5b4947c8bfc5b'
__git_head_subject__ = 'Finished release v0.6.15.'
__git_head_message__ = ''
__git_dirty_diff__ = ''